function   [ fulldata_model, fulldata_misc, summary_model, summary_misc ] = fit_1d1g1sr( in_data, par )

% function [ fulldata_model, fulldata_misc, summary_model, summary_misc ] = fit_1d1g1sr( data, par )
%
% Fit a 1d1g1sr model to some 2-D data.
%
%
% INPUT (data only is mandatory)
%
% data: a 2-D matrix of >= 0 values
%
% par:  a structure where each field defines a parameter
%
%
% OUTPUT
%
% model is a structure with fields  
% { 'b0', 'w0d', 'gamma0', 'beta0', 'sigma1', 'V1' }.
%
% ( Implicitly:  b1 = 1 - b0, w0g = 1 - w0d, and the shift is gamma0 * beta0 )
%
% misc is a structure with fields { 'alldata_log_plkld', 'iter' }
%
%
% SEE ALSO compute_post_1d1g1sr, fit_1d1g1sr

  DEBUG_HIST_NBINS = 100;
  
  if nargin < 1
    error( [ mfilename ' needs at least one input argument.' ] );
  end

  if any( isnan( in_data(:) ) )
    error( [ mfilename ': needs non-NaN data!' ] );
  end
  
  if ~exist( 'par', 'var' )
    par = [];
  end

  % data: save some memory
  global summary fulldata   negloglike_w0   negloglike_w1  GAMMA0_FLOOR  BETA0_FLOOR;
  
  % This is not supposed to be changed
  GAMMA0_FLOOR = 1 + 1e-10;
  BETA0_FLOOR  = 1e-10;
  

  fulldata.data = in_data; clear in_data;
  
  % data: make sure it is a 2-D matrix
  
  size_data = size( fulldata.data );
  
  if max( find( size_data > 1 ) ) > 2
    error( [ mfilename ': "data" must be a 2-D matrix.' ] );
  end
  
  % Now we can quietly drop its size
  
  fulldata.data = fulldata.data(:);
  fulldata.numel_data = numel( fulldata.data );
  
  % Check all data is >= 0
  
  if any( fulldata.data(:) < 0 )
    error( [ mfilename ': all data must be >=0.' ] );
  end
  
  fulldata.z_ind  = find( fulldata.data == 0 );
  fulldata.z_ind  = fulldata.z_ind(:);
  
  fulldata.nz_ind = find( fulldata.data > 0 );
  fulldata.nz_ind = fulldata.nz_ind(:);
  
  fulldata.log_data_nz_ind = log( fulldata.data( fulldata.nz_ind ) );
  fulldata.log_data_nz_ind = fulldata.log_data_nz_ind(:);
  
  % Check parameters, fill in default values
  
  par_default = [];
  
  % model_init:
  % - if empty, automatic initialization is used.
  % - if non-empty, this field contains a user-defined 1d1g1sr model to initialize with.
  par_default.model_init = [];
  
  % Alternative: automatic initialization
  % We try several splits of the data, 
  % apply a rough moment-based initialization for each,
  % and pick the ML initialization.
  par_default.auto_init_n = 30;
  
  % How many frames to retain in the summarized version of the data
  par_default.summary_nsamples = min( 1000, fulldata.numel_data );
  
  
  par_default.summary_min_iter = 0;
  par_default.summary_max_iter = +Inf;
  
  % To accelerate convergence ( >=0 value or [] to desactivate it).
  % Danger: we may converge to a suboptimal maxima,
  % as compared to the basic search (inertia = []).
  par_default.summary_inertia  = [];
  
  % Convergence detected when alldata_log_plkld varied by less than this ...
  par_default.summary_lkld_conv = 1e-10;
  % ...AND when parameters change by less than this
  par_default.param_conv        = 1e-5;
  
  par_default.fulldata_min_iter = 0;
  par_default.fulldata_max_iter = +Inf;
  
  % To accelerate convergence ( >=0 value or [] to desactivate it).
  % Danger: we may converge to a suboptimal maxima,
  % as compared to the basic search (inertia = []).
  par_default.fulldata_inertia  = [];
  
  % Convergence detected when alldata_log_plkld varied by less than this
  par_default.fulldata_lkld_conv = 1e-10;
  
  % 0: moment method. 
  %    - advantage: fast (analytical).
  %    - disadvan.: ML not guaranteed.
  %
  % 1: ML numerical optimization.
  par_default.do_fminsearch = 1;
  
  par_default.fminsearch_options = [];
  
  % To make sure the weight of Dirac never falls to zero
  % This is necessary if all of the training data is > 0,
  % but some testing data is = 0.
  par_default.wd_floor = 1e-5;

  % To make sure the Rice is correctly defined
  par_default.sigma_floor = 1e-10;
  
  par_default.verbose = 1;
  par_default.debug   = 0;

  % Title name to put in the figures
  % ( used only if par.debug = 1 )
  par_default.debug_title_name = datestr( clock );
  
  % With this one you can choose between
  % slightly faster (0) or safe (1)
  par_default.sanity_check = 1;
  
  par = fill_default( par, par_default );
  
  if par.verbose
    disp( ' ' );
    disp( [ mfilename ' parameters:' ] );
    disp( par );
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( 1 ) Initialization
  
  
  % Create a summarized version of the data
  
  sort_data = sort( fulldata.data );
  
  summary.data = sort_data( round( ( 0.5:par.summary_nsamples ) / par.summary_nsamples * fulldata.numel_data ) );
  summary.numel_data = numel( summary.data );
  summary.nz_ind = find( summary.data(:)  > 0 );
  summary.z_ind  = find( summary.data(:) == 0 );
  summary.log_data_nz_ind = log( summary.data( summary.nz_ind ) );
  
  % Sanity check

  if any( summary.numel_data ~= par.summary_nsamples )
    error( [ mfilename ': insanity detected while creating the summarized data.' ] );
  end
  
  % Initialize the model
  
  if ~isempty( par.model_init )

    % User-defined init

    if par.verbose
      disp( [ mfilename ': user-defined initialization.' ] );
    end
    
    % Check model: make sure all parameters are defined
    a_new.model = check_1d1g1sr_model( par.model_init );
    
    
  else

    if par.verbose
      disp( [ mfilename ': automatic initialization.' ] );
	    
    end
    
    % Default, automatic init
    
    % Initialize the distributions
    
    
    sort_data = sort( summary.data( summary.nz_ind ) );
    
    if par.auto_init_n < 2 
      
      % Single initialization
      
      % Equal priors
      a_new.model.b0   = 0.5;

      % Update mixture weights
      a_new.model.w0d = min( 0.9, max( par.wd_floor, mean( summary.data == 0 ) / a_new.model.b0 ) );

      % Define threshold
      thr_list = sort_data( max( 2, round( a_new.model.b0 * summary.numel_data ) ) );
      
    else
      
      % Multiple initializations
      % (we'll pick the ML one)
      % -> this way we avoid gross mistakes
      %    that would lead to bad local optima
      
      % About half of the list of thresholds are 
      % regularly spaced between a_min and a_max.
      
      a_max = sort_data( end );
      a_min = sort_data( 1 );

      n1 = max( 1, round( par.auto_init_n / 2 ) );
      
      thr_list = a_min + (0.5:n1) / n1 * ( a_max - a_min );
      
      % The other half of the list of thresholds are
      % regularly spaced along percentiles.
      
      n2 = max( 0, par.auto_init_n - n1 );
      
      n = numel( sort_data );
      ind = 1 + round( ( 0.5:n2 ) / n2 * ( n - 1 ) );
      
      tmp = sort_data( ind );
      
      thr_list = [ thr_list(:).' tmp(:).' ];
      
    end
      
    thr_list = sort( thr_list );
    
    for a = 1:numel( thr_list )
    
      thr = thr_list( a );
      
      % Gamma pdf for silence: moment method
      
      lsr_ind = find( sort_data <= thr );
      a_mean = mean( sort_data( lsr_ind ) );
      a_var  = var( sort_data( lsr_ind ) );
      
      a_new.model.gamma0 = a_mean ^ 2 / a_var;
      a_new.model.beta0  = a_mean / a_new.model.gamma0;
      
      % Ensure that GammaPdf( 0, gamma0, beta0 ) = 0
      
      t0 = a_new.model.gamma0 * a_new.model.beta0;
      
      a_new.model.gamma0 = max( GAMMA0_FLOOR, a_new.model.gamma0 );
      a_new.model.beta0  = t0 / a_new.model.gamma0;
      
      % Shifted Rice pdf for activity: moment method
      
      a_shift = shift_1d1g1sr( a_new.model.gamma0, a_new.model.beta0 );
      grt_ind = find( sort_data > max( a_shift, thr ) );
      
      if numel( grt_ind ) < 2
	
	% could not find any data to initialize the Shifted Rice
	log_plkld_list( a ) = -Inf;
      
      else
	
	[ a_new.model.sigma1, a_new.model.V1 ] = rice_optimize_sigma_V( sort_data( grt_ind ) - a_shift );
	a_new.model.sigma1 = max( par.sigma_floor, a_new.model.sigma1 );
	
	% We are almost sure to get those crappy warnings
	% so let's shut if off during initilization
	old_warning_state = warning;
	warning off;
	
	% Update priors
	
	a_new.model.b0 = min( 0.9, max( 1e-1, mean( summary.data <= thr ) ) );
	
	% Update mixture weights
	
	a_new.model.w0d = min( 0.9, max( par.wd_floor, mean( summary.data == 0 ) / a_new.model.b0 ) );
	
	% Compute lkld
	
	
	try	
	  [ tmp, log_plkld_list( a ) ] = compute_post_1d1g1sr( a_new.model, summary.data, 1 );
	catch
	  % compute_post_1d1g1sr failed, mostly because of saturation of the besseli(0,x) function in the Rice
	  log_plkld_list( a ) = -Inf;
	end
      
      end
      
      warning( old_warning_state );
      
      model_list( a ) = a_new;

      % Back to the old warning state
      warning( old_warning_state );
      
    end
    
    % Pick the ML
    
    [ tmp, max_ind ] = max( log_plkld_list );
    
    a_new = model_list( max_ind );
    
    if par.debug
      figure; plot( thr_list,log_plkld_list,'.-' )
      xlabel( 'init threshold' ); ylabel( 'alldata log lkld' );
      hold on; plot( thr_list( max_ind ), log_plkld_list( max_ind ), 'ro' );
      title( strrep( [ mfilename ' initialization - ' par.debug_title_name ], '_', '\_' ) );

    end
    
  end  % of "if automatic init"
  
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( 2 ) EM training
  % First on the summarized version of the data
  % Second on the full data
  
  summary.best  = [];
  fulldata.best = [];
  
  summary.iter = 0;
  summary.finished = ( summary.iter >= par.summary_max_iter );
  summary.previous_alldata_log_plkld = [];
  summary.previous_X0 = [];
  
  fulldata.iter = -1;
  fulldata.finished = ( fulldata.iter >= par.fulldata_max_iter );
  fulldata.previous_alldata_log_plkld = [];
  fulldata.previous_X0 = [];
  
  % Special case: start directly with the full data
  if par.summary_max_iter < 1
    fulldata.iter = 0;
  end
  
  % For convergence test
  previous_params = [];
  param_delta = [];
  
  while ~( summary.finished & fulldata.finished )
    
    if par.verbose
      tic
    end
    
    % E step: estimate posteriors
    
    if ~summary.finished
      [ sector_log_post, a_new.alldata_log_plkld ] = compute_post_1d1g1sr( a_new.model, summary.data, par.sanity_check );
    else
      [ sector_log_post, a_new.alldata_log_plkld ] = compute_post_1d1g1sr( a_new.model, fulldata.data, par.sanity_check );
    end
    
    % Check whether or not the parameters are converging
    
    new_params = [ a_new.model.b0  a_new.model.w0d  a_new.model.gamma0  a_new.model.beta0  a_new.model.sigma1  a_new.model.V1 ];
    param_delta = [];
    is_converging = 0;
    if ~isempty( previous_params )
      param_delta = max( abs( new_params - previous_params ) );
      is_converging =  ( param_delta < par.param_conv );
    end
    previous_params = new_params;
    
    
    % Verbosity
    
    if par.verbose
      
      disp( ' ' );
      
      if ~summary.finished
	disp( sprintf( [ mfilename ': model at iteration summary.iter:%d' ], summary.iter ) );
      else
	disp( sprintf( [ mfilename ': model at iteration fulldata.iter:%d' ], fulldata.iter ) );
      end
      
      disp( a_new.model );
      
      if ~summary.finished
	previous_alldata_log_plkld = summary.previous_alldata_log_plkld;
      else
	previous_alldata_log_plkld = fulldata.previous_alldata_log_plkld;
      end
      
      s = sprintf( [ mfilename ': alldata_log_plkld:%g' ], a_new.alldata_log_plkld );
      if ~isempty( previous_alldata_log_plkld )
	s = [ s  sprintf( ' (lkld_delta:%g)', a_new.alldata_log_plkld - previous_alldata_log_plkld ) ];
      end
      if ~isempty( param_delta )
	s = [ s sprintf( ' (param_delta:%g)', param_delta ) ];
      end
      disp( s );
      
      if ~summary.finished
	summary.previous_alldata_log_plkld = a_new.alldata_log_plkld;
      else
	fulldata.previous_alldata_log_plkld = a_new.alldata_log_plkld;
      end
      
    end
    
    
    % Check whether or not the likelihood is converging
    if ( ~summary.finished ) & ( ~isempty( summary.best ) ) & ( ~isempty( summary.previous_alldata_log_plkld ) )
      is_converging = is_converging & ( summary.best.alldata_log_plkld > a_new.alldata_log_plkld - par.summary_lkld_conv );
    elseif ( ~isempty( fulldata.best ) ) & ( ~isempty( fulldata.previous_alldata_log_plkld ) )
      is_converging = is_converging & ( fulldata.best.alldata_log_plkld > a_new.alldata_log_plkld - par.fulldata_lkld_conv );
    end
      
    
    
    % Look for improvement
    
    % Always update in some particular cases
    is_improving = ( ( ~summary.finished ) & isempty( summary.best ) ) | ( ( summary.finished ) & isempty( fulldata.best ) );

    if ~is_improving
    
      % General cases: compare partial likelihoods
      if ~summary.finished
	is_improving = (  summary.best.alldata_log_plkld < a_new.alldata_log_plkld );
      else
	is_improving = ( fulldata.best.alldata_log_plkld < a_new.alldata_log_plkld );
      end
      
    end
    
    % If improvement, then update the "best so far"
    
    if is_improving
      
      if ~summary.finished
	summary.best = a_new;
      else
	fulldata.best = a_new;
      end
      
    end
    
    % Termination test
    if ~summary.finished
      summary.finished  = ( is_converging & ( summary.iter >= par.summary_min_iter ) ) | ( summary.iter >= par.summary_max_iter );
      if summary.finished
	
	previous_params = [];
	param_delta = [];
      end
    else
      fulldata.finished = ( is_converging & ( fulldata.iter >= par.fulldata_min_iter ) ) | ( fulldata.iter >= par.fulldata_max_iter );
    end
    
    % M step: update parameters
    
    if ~( summary.finished & fulldata.finished )

      if summary.finished & ( fulldata.iter == -1 )
	
	% Transition from the summarized data set to the full data set
	% -> pick the best model so far
	
	model = summary.best.model;
	
      else
	
	% General case: regular update
	% Update priors of silent sector and active sector
	tmp = my_logsum_fast( reshape( sector_log_post, [], 2 ) );
	tmp = exp( tmp - my_logsum_fast( tmp ) ); % Normalization
	a_new.model.b0 = tmp( 1 ) / sum( tmp );  % Normalization for minor numerical errors
	
	% Update weights of the ( silent sector ) mixture
	tmp_mat = sector_log_post( :,:,1 );
	if ~summary.finished
	  tmp      = my_logsum_fast( reshape( tmp_mat( summary.z_ind  ), [], 1 ) );
	  tmp( 2 ) = my_logsum_fast( reshape( tmp_mat( summary.nz_ind ), [], 1 ) );
	else
	  tmp      = my_logsum_fast( reshape( tmp_mat( fulldata.z_ind  ), [], 1 ) );
	  tmp( 2 ) = my_logsum_fast( reshape( tmp_mat( fulldata.nz_ind ), [], 1 ) );
	end
	tmp = exp( tmp - my_logsum_fast( tmp ) );% Normalization
	a_new.model.w0d = tmp( 1 ) / sum( tmp );  % Normalization for minor numerical errors
	a_new.model.w0d = max( par.wd_floor, a_new.model.w0d ); % Flooring
	
	%%%%%%%%%%
	
	% Update the ( gamma0, beta0, sigma1, V1 ) parameters

	t0 = a_new.model.gamma0 * a_new.model.beta0;
	
	X0 = [ t0;
	       a_new.model.gamma0;   
	       a_new.model.sigma1;
	       a_new.model.V1; ];
	
	negloglike_w0   = exp( sector_log_post( :,:,1 ) );
	negloglike_w1   = exp( sector_log_post( :,:,2 ) );
	
	fval0 = negloglike( X0, par.sigma_floor );
	
	
	% Optimization of the parameters: either moment-based or numerical search
	
	if par.do_fminsearch == 0
	  
	  % Moment method
	  % ( fast, analytical, but NOT ML )
	  
	  if ~summary.finished

	    % Gamma
	    
	    tmp_sum = sum( negloglike_w0( summary.nz_ind(:) ) );
	    a_mean = negloglike_w0( summary.nz_ind(:) ).' * summary.data( summary.nz_ind(:) ) / tmp_sum;
	    a_var  = negloglike_w0( summary.nz_ind(:) ).' * ( ( summary.data( summary.nz_ind(:) ) - a_mean ) .^ 2 ) / tmp_sum;
	    
	    X( 1 ) = a_mean;              % gamma0 * beta0
	    X( 2 ) = a_mean ^ 2 / a_var;  % gamma0
	    
	    % Ensure that GammaPdf( 0, gamma0, beta0 ) = 0
	    
	    X( 2 ) = max( GAMMA0_FLOOR, X( 2 ) );
	    	    
	    % Shifted Rice
	    
	    a_shift = shift_1d1g1sr( X( 2 ), X( 1 ) / X( 2 ) );
	    grt_ind = find( summary.data > a_shift );
	    if numel( grt_ind ) < 2
	      error( [ mfilename ' could not find any data to fit the Shifted Rice on (moment, summary).' ] );
	    end
	    
	    [ X( 3 ), X( 4 ) ] = rice_optimize_sigma_V( summary.data( grt_ind ) - a_shift, negloglike_w1( grt_ind ) );
	    
	  else
	    
	    % Gamma
	    
	    tmp_sum = sum( negloglike_w0( fulldata.nz_ind(:) ) );
	    a_mean = negloglike_w0( fulldata.nz_ind(:) ).' * fulldata.data( fulldata.nz_ind(:) ) / tmp_sum;
	    a_var  = negloglike_w0( fulldata.nz_ind(:) ).' * ( ( fulldata.data( fulldata.nz_ind(:) ) - a_mean ) .^ 2 ) / tmp_sum;
	    
	    X( 1 ) = a_mean;              % gamma0 * beta0
	    X( 2 ) = a_mean ^ 2 / a_var;  % gamma0

	    % Ensure that GammaPdf( 0, gamma0, beta0 ) = 0
	    
	    X( 2 ) = max( GAMMA0_FLOOR, X( 2 ) );
	    	    
	    % Shifted Rice
	    
	    a_shift = shift_1d1g1sr( X( 2 ), X( 1 ) / X( 2 ) );
	    grt_ind = find( fulldata.data > a_shift );
	    if numel( grt_ind ) < 2
	      error( [ mfilename ' could not find any data to fit the Shifted Rice on (moment, fulldata).' ] );
	    end
	    
	    [ X( 3 ), X( 4 ) ] = rice_optimize_sigma_V( fulldata.data( grt_ind ) - a_shift, negloglike_w1( grt_ind ) );
	    
	  end
	  
	  % Flooring
	  X( 3 ) = max( par.sigma_floor, X( 3 ) );
	  
	  % Evaluation
	  fval = negloglike( X, par.sigma_floor );
	  
	else
	  
	  % ML (numerical optimization)
	
	  % Optional inertia in the variation of the parameters
	  
	  if ~summary.finished
	    
	    tmp = X0;
	    
	    if ~( isempty( par.summary_inertia ) | isempty( summary.previous_X0 ) )
	      X0 = max( [ GAMMA0_FLOOR * BETA0_FLOOR; GAMMA0_FLOOR; par.sigma_floor; 0 ], X0 + par.summary_inertia * ( X0 - summary.previous_X0 ) );
	    end
	    
	    summary.previous_X0 = tmp;
	    
	  elseif ~fulldata.finished
	    
	    tmp = X0;
	    
	    if ~( isempty( par.fulldata_inertia ) | isempty( fulldata.previous_X0 ) )
	      X0 =  max( [ GAMMA0_FLOOR * BETA0_FLOOR; GAMMA0_FLOOR; par.sigma_floor; 0 ], X0 + par.fulldata_inertia * ( X0 - fulldata.previous_X0 ) );
	    end
	    
	    fulldata.previous_X0 = tmp;
	    
	  end

	  if par.sanity_check
	    % For the sanity check
	    tmp_fval = negloglike( X0, par.sigma_floor );
	  end
	  
	  [ X, fval ] = fminsearch( @negloglike, X0, par.fminsearch_options, par.sigma_floor );
	
	  if par.sanity_check
	    % Sanity check
	    if -fval < -tmp_fval
	      error( [ mfilename ' decreased the likelihood during ( gamma0, beta0 ) optimization!' ] );
	    end
	  end
	  
	end

	% In both "moment" and "ML" cases,
	% update only if we have an improvement.
	% This way we guarantee to always increase the lkld.
	if -fval > -fval0
	  a_new.model.gamma0  = X( 2 );
	  a_new.model.beta0   = X( 1 ) / X( 2 );
	  a_new.model.sigma1  = X( 3 );
	  a_new.model.V1      = X( 4 );
	end

      end      
      
    end
    
    
    % Prepare for the next iteration
    if ~summary.finished
      summary.iter = summary.iter + 1;
    elseif ~fulldata.finished
      fulldata.iter = fulldata.iter + 1;
    end
    
    if par.verbose
      disp( ' ' );
      disp( sprintf( [ mfilename ': iter duration was %g sec' ], toc ) );
    end
  end  % of "while" loop

  % Verbosity to conclude the training
  
  if par.verbose
    disp( ' ' );
    disp( sprintf( [ mfilename ': training finished.' ] ) )

    disp( sprintf( [ '  summary.iter:%d, summary.best.alldata_log_plkld:%g, summary.best.model:' ], summary.iter, summary.best.alldata_log_plkld ) );
    disp( summary.best.model );
    
    disp( sprintf( [ '  fulldata.iter:%d, fulldata.best.alldata_log_plkld:%g, fulldata.best.model:' ], fulldata.iter, fulldata.best.alldata_log_plkld ) );
    disp( fulldata.best.model );

  end
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( 3 ) Return values
  
  fulldata_model = [];
  fulldata_misc.alldata_log_plkld = [];
  if ~isempty( fulldata.best )
    fulldata_model = fulldata.best.model;
    fulldata_misc.alldata_log_plkld = fulldata.best.alldata_log_plkld;
  end
  fulldata_misc.iter = fulldata.iter;
  
  summary_model = [];
  summary_misc.alldata_log_plkld = [];
  if ~isempty( summary.best )
    summary_model = summary.best.model;
    summary_misc.alldata_log_plkld = summary.best.alldata_log_plkld;
  end
  summary_misc.iter = summary.iter;
  
  %%%
  
  if par.debug
    
    %%%%%%%%%%
    % Model trained on summarized data
    
    if ~isempty( summary.best )
      model = summary.best.model;
      
      % Show the distributions
      
      [ h.w, h.x ] = hist( fulldata.data(:), DEBUG_HIST_NBINS );
      h.x = h.x(:);
      h.w = running_mean( h.w, 3 );
      h.w = h.w(:) / sum( h.w );
      
      [ h.sector_log_post, h.alldata_log_plkld, h.sector_log_plkld, h.sector_log_pjoint ] = compute_post_1d1g1sr( model, h.x );
      
      a_mat = exp( [ permute( h.sector_log_pjoint, [1 3 2] ) ] );
      a_mat = [ a_mat  sum( a_mat, 2 ) ];
      a_mat = a_mat / sum( a_mat(2:end,end) ) * sum( h.w(2:end) );
      
      a_mat( 1,1 ) = a_mat( 1,1 ) + model.w0d;
      
      figure; set( bar( h.x, h.w ), 'edgecolor', [ 0.7 0.7 0.7 ], 'facecolor', [ 0.7 0.7 0.7 ] );
      hold on; plot( h.x, a_mat( :,1:end-1), 'linewidth', 2 );
      plot( h.x, a_mat( :,end ), 'c--', 'linewidth', 2 );
      
      xlabel( 'data' );
      title( strrep( [ 'summary - ' par.debug_title_name ], '_', '\_' ) );
      
      % Show the distributions in log domain
      
      [ hlog.w, hlog.x ] = hist( log( fulldata.data( fulldata.nz_ind ) ), DEBUG_HIST_NBINS );
      hlog.x = hlog.x(:);
      hlog.w = running_mean( hlog.w, 3 );
      hlog.w = hlog.w(:) / sum( hlog.w );
      
      [ hlog.sector_log_post, hlog.alldata_log_plkld, hlog.sector_log_plkld, hlog.sector_log_pjoint ] = compute_post_1d1g1sr( model, exp( hlog.x ) );
      
      a_mat = exp( [ permute( hlog.sector_log_pjoint, [1 3 2] ) ] );
      a_mat = diag( sparse( exp( hlog.x ) ) ) * [ a_mat  sum( a_mat, 2 ) ];
      a_mat = a_mat / sum( a_mat(2:end,end) ) * sum( hlog.w(2:end) );
      
      a_mat( 1,1 ) = a_mat( 1,1 ) + model.w0d;
      
      figure; set( bar( hlog.x, hlog.w ), 'edgecolor', [ 0.7 0.7 0.7 ], 'facecolor', [ 0.7 0.7 0.7 ] );
      hold on; plot( hlog.x, a_mat( :,1:end-1), 'linewidth', 2 );
      plot( hlog.x, a_mat( :,end ), 'c--', 'linewidth', 2 );
      
      xlabel( 'log( data )' );
      title( strrep( [ 'summary - ' par.debug_title_name ], '_', '\_' ) );
      
    end

    %%%%%%%%%%
    % Full-data model
    
    if ~isempty( fulldata.best )
      model = fulldata.best.model;
      
      % Show the distributions
      
      [ h.w, h.x ] = hist( fulldata.data(:), DEBUG_HIST_NBINS );
      h.x = h.x(:);
      h.w = running_mean( h.w, 3 );
      h.w = h.w(:) / sum( h.w );
      
      [ h.sector_log_post, h.alldata_log_plkld, h.sector_log_plkld, h.sector_log_pjoint ] = compute_post_1d1g1sr( model, h.x );
      
      a_mat = exp( [ permute( h.sector_log_pjoint, [1 3 2] ) ] );
      a_mat = [ a_mat  sum( a_mat, 2 ) ];
      a_mat = a_mat / sum( a_mat(2:end,end) ) * sum( h.w(2:end) );
      
      a_mat( 1,1 ) = a_mat( 1,1 ) + model.w0d;
      
      figure; set( bar( h.x, h.w ), 'edgecolor', [ 0.7 0.7 0.7 ], 'facecolor', [ 0.7 0.7 0.7 ] );
      hold on; plot( h.x, a_mat( :,1:end-1), 'linewidth', 2 );
      plot( h.x, a_mat( :,end ), 'c--', 'linewidth', 2 );
      
      xlabel( 'data' );
      title( strrep( [ 'fulldata - ' par.debug_title_name ], '_', '\_' ) );
      
      % Show the distributions in log domain
      
      [ hlog.w, hlog.x ] = hist( log( fulldata.data( fulldata.nz_ind ) ), DEBUG_HIST_NBINS );
      hlog.x = hlog.x(:);
      hlog.w = running_mean( hlog.w, 3 );
      hlog.w = hlog.w(:) / sum( hlog.w );
      
      [ hlog.sector_log_post, hlog.alldata_log_plkld, hlog.sector_log_plkld, hlog.sector_log_pjoint ] = compute_post_1d1g1sr( model, exp( hlog.x ) );
      
      a_mat = exp( [ permute( hlog.sector_log_pjoint, [1 3 2] ) ] );
      a_mat = diag( sparse( exp( hlog.x ) ) ) * [ a_mat  sum( a_mat, 2 ) ];
      a_mat = a_mat / sum( a_mat(2:end,end) ) * sum( hlog.w(2:end) );
      
      a_mat( 1,1 ) = a_mat( 1,1 ) + model.w0d;
      
      figure; set( bar( hlog.x, hlog.w ), 'edgecolor', [ 0.7 0.7 0.7 ], 'facecolor', [ 0.7 0.7 0.7 ] );
      hold on; plot( hlog.x, a_mat( :,1:end-1), 'linewidth', 2 );
      plot( hlog.x, a_mat( :,end ), 'c--', 'linewidth', 2 );
      
      xlabel( 'log( data )' );
      title( strrep( [ 'fulldata - ' par.debug_title_name ], '_', '\_' ) );
      
    end

  end
  


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function y = negloglike( X, sigma_floor )
  
% Evaluate the negative of the Gamma + Shifted Rice terms in the log lkld
% The rest of the terms in the log lkld are constants 
% w.r.t. gamma0, beta0, sigma1, V1.
  
  if nargin < 2
    error( [ mfilename '.negloglike() needs 2 input arguments.' ] );
  end
  
  % "global" variables are used to transmit data arrays without useless memory waste
  
  global summary  fulldata  negloglike_w0  negloglike_w1  GAMMA0_FLOOR  BETA0_FLOOR;
  
  % Get the parameters
  
  gamma0 = X( 2 );
  beta0  = X( 1 ) / X( 2 );
  sigma1 = X( 3 );
  V1     = X( 4 );
  
  % Impose constraints on them
  
  if ( gamma0 < GAMMA0_FLOOR ) | ( beta0 < BETA0_FLOOR ) | ( sigma1 < sigma_floor ) | ( V1 < 0 )
    y = +Inf; return;
  end
  
  aV1 = abs( V1 );
  
  % Do it!
  
  y = 0;
  
  if ~summary.finished

    % Gamma
    
    y = y - negloglike_w0( summary.nz_ind(:) ).' * ( ( gamma0 - 1 ) .* summary.log_data_nz_ind(:) - summary.data( summary.nz_ind(:) ) / beta0 - gammaln( gamma0 ) - gamma0 * log( beta0 ) );
    
    % Shifted Rice
    a_shift = shift_1d1g1sr( gamma0, beta0 );
    grt_ind = find( summary.data > a_shift );
    if ~isempty( grt_ind )
      tmp = summary.data( grt_ind(:) ) - a_shift;
      y = y - negloglike_w1( grt_ind(:) ).' * ( log( tmp(:) ) - 2 * log( sigma1 ) - ( tmp(:) .^ 2 + aV1 ^ 2 ) / ( 2 * sigma1 ^ 2 ) + log( besseli( 0, tmp * aV1 / sigma1 ^ 2 ) ) );
    end
    
  else
    
    % Gamma
    y = y - negloglike_w0( fulldata.nz_ind(:) ).' * ( ( gamma0 - 1 ) .* fulldata.log_data_nz_ind(:) - fulldata.data( fulldata.nz_ind(:) ) / beta0 - gammaln( gamma0 ) - gamma0 * log( beta0 ) );
    
    % Shifted Rice
    a_shift = shift_1d1g1sr( gamma0, beta0 );
    grt_ind = find( fulldata.data > a_shift );
    if ~isempty( grt_ind )
      tmp = fulldata.data( grt_ind(:) ) - a_shift;
      y = y - negloglike_w1( grt_ind(:) ).' * ( log( tmp(:) ) - 2 * log( sigma1 ) - ( tmp(:) .^ 2 + aV1 ^ 2 ) / ( 2 * sigma1 ^ 2 ) + log( besseli( 0, tmp * aV1 / sigma1 ^ 2 ) ) );
    end
    
  end
  
  % Normalize y by its number of terms,
  % so that "TolFun" may have a meaning
  
  if ~summary.finished
    y = y / numel( summary.data(:) );
  else
    y = y / numel( fulldata.data(:) );
  end
  