function s060720_twoarrays_onecam_bgsubgmm( p )
  
  if nargin < 1
    error( [ mfilename ' needs 1 input argument.' ] );
  end

  required_fields = { 'input_template', 'videoframerate', 'input_timecodefile', 'input_classif_namehead', 'input_arrayfile_list', 'audiostart_videotime', 'output_namehead' };
  check_param( required_fields, fieldnames( p ) );

  dbstop error

  % Check whether the output model file already exists
  
  output_model_filename = [ p.output_namehead '_' mfilename '_last_models.mat' ];

  if exist( output_model_filename, 'file' )

    disp( [ mfilename ': output file already exists. Skipped.' ] );
    disp( [ '"' output_model_filename '"' ] );

    return;

  end

  % If not, lock it by creating it

  % Make sure that the output directory exists
  output_path = fileparts( output_model_filename );
  if ~exist( output_path, 'dir' )
    system( [ 'mkdir -p ' output_path ] );
    if p.verbose
      disp( ' ' );
      disp( [ mfilename ' created output directory "' output_path '".' ] );
    end
  end
  
  % Create an empty file (useful when processing many recordings on several machines)
  fid = fopen( output_model_filename, 'wt' );
  fclose( fid );
  
  % Look at the raw images
  
  d = dir( p.input_template );
  d = d( find( ~[ d.isdir ] ) );
  
  input_path = fileparts( p.input_template )
  name_list  = sort( { d.name } );
  
  % Also load their timecodes
  
  timecode_list = load_timecodes( p.input_timecodefile, p.videoframerate );
  
  % Look only at images for which we have a timecode
  
  [ name_list, ia, ib ] = intersect( name_list, timecode_list( 1, : ) );
  
  % Store the list of timecodes as a vector array of scalars (time values in seconds).
  
  timecode_list = [ timecode_list{ 2, ib } ];
  
  n_image = numel( name_list );
  
  disp( sprintf( [ mfilename ': %d image files \n(input_path: "%s")\n(files:      "%s" to "%s")\n' ], n_image, input_path, name_list{ [ 1 end ] } ) );

  % Make sure that we also have the original background subtraction result, for each image

  for a = 1:n_image
    
    a_filename = [ p.input_classif_namehead name_list{ a } ];
    if ~exist( a_filename, 'file' )
      disp( [ mfilename ': could not find original bg sub result for image "' name_list{ a } '".' ] );
      disp( [ mfilename ': missing file: "' a_filename '".' ] );
      return;
    end

  end
  
  % For each video frame, load the corresponding audio location estimates.
  % 
  % audioloc_list is a 3 x n_audioloc array
  %   row 1: video time in seconds
  %   row 2: azimuth in radians
  %   row 3: elevation in radians

  n_array = numel( p.input_arrayfile_list )
  
  if n_array > 8
    disp( [ mfilename ': cannot support more than 8 arrays.' ] );
    disp( [ mfilename ': you just have to change a small part of me: the way speechpix_mat is encoded.' ] );
  end
  
  for a = 1:n_array

    array( a ).audioloc_mat = load_audio( p.input_arrayfile_list{ a }, p.audiostart_videotime( a ) );

    disp( [ mfilename ': loaded audio location estimates.' ] );
    disp( [ mfilename ': from file "' p.input_arrayfile_list{ a } '".' ] );
    disp( ' ' );
    
    % Where we'll store the A|V model

    array( a ).av_model = [];

    array( a ).par = p;

  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( 1 ) Define parameters
  
  if ~exist( 'p', 'var' )
    p = [];
  end

  clear p_default;

  % Possible mask to multiply the image with
  % ( default: none, as indicated by the [] parameter )
  
  p_default.mask = [];

  %%% ( 1.1 ) ( A | V ) models

  % "Major" parameters (according to the PAMI article)

  p_default.a_alpha = 1e-1;
  p_default.a_T     = 0.66;
  
  % "Minor" parameters (according to the PAMI article)
  
  p_default.a_K          = 3;
  p_default.a_Z          = 2.5;
  p_default.a_sigma_init = [ 10/180*pi; 30/180 * pi ];
  p_default.a_w_init     = 0.01;
  
  p_default.a_use_fixed_rho = 0;  % Binary value
  p_default.a_rho_factor    = 1e-5;  % If "use_fixed_rho == 1" then rho = rho_factor * alpha

  p_default.a_sigma_floor  = 1e-3;

  %%% ( 1.2 ) Improved V models

  % "Major" parameters (according to the PAMI article)

  p_default.v_alpha = 1e-2;
  p_default.v_T     = 0.66;
  
  % "Minor" parameters (according to the PAMI article)
  
  p_default.v_K          = 3;
  p_default.v_Z          = 2.5;
  p_default.v_sigma_init = 20;
  p_default.v_w_init     = 0.01;
  
  p_default.v_use_fixed_rho = 0;  % Binary value
  p_default.v_rho_factor    = 1e-5;  % If "use_fixed_rho == 1" then rho = rho_factor * alpha
  
  %%% ( 1.3 ) Misc.

  p_default.verbose = 1;
  
  %%% Fill up missing parameters
  
  p = fill_default( p, p_default );

  if p.verbose
    disp( [ mfilename ' parameters:' ] );
    disp( p );
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( 2 ) Process frame by frame
  
  % Where we'll store the improved video background model

  clear v_model
  v_model = [];


  if p.verbose
    chrono = chrono_start( n_image, 30 );
  end

  for a = 1:n_image

    disp( sprintf( 'video: %d / %d', a, n_image ) ); % XXX DEBUG
    
    % ( 2.1 ) load the image

    im = Lecture( fullfile( input_path, name_list{ a } ) );

    % Optional masking

    if ~isempty( p.mask )
      im = im .* p.mask;
    end

    % Note that the image is usually RGB, insert possible HSV transform here

    im_size = size( im );
    
    if numel( im_size ) == 2
      im_size( 3 ) = 1;
    end

    if numel( im_size) ~= 3
      error( [ mfilename ': insanity detected, concerning "im_size".' ] );
    end

    % ( 2.2 ) Load the original bg sub result

    bg = LecturePgm( [ p.input_classif_namehead  name_list{ a } ] );

    
    % ( 2.3 ) Update the ( A | V ) model, in ( az, el ) space
    
    % Where we will store the speech pixels

    speechpix_mat = zeros( im_size( 1:2 ) );

    % First we need to list the relevant audio location estimates
    
    videotime = timecode_list( a );

    for i_array = 1:n_array
      
      % XXX DEBUG
      disp( sprintf( '  array %d / %d', i_array, n_array ) );

      loc_ind   = find( abs( videotime - array( i_array ).audioloc_mat( 1, : ) ) < 0.5 / p.videoframerate );
      
      azel_mat = array( i_array ).audioloc_mat( 2:3, loc_ind );
      
      % Now we can update the ( A | V ) model for each foreground pixel
      
      if ( ~isempty( azel_mat ) ) & ( any( bg(:) > 0 ) )
	
	array( i_array ).av_model = update_av_model( array( i_array ).av_model, azel_mat, bg, im_size, p ); 
	
      end
      
      % ( 2.4 ) List pixels that are matched by the current A measurements

      if ( ~isempty( azel_mat ) ) & ( ~isempty( array( i_array ).av_model ) )

	speechpix_mat = speechpix_mat + ( 2 ^ ( i_array - 1 ) ) * detect_speech_pixels( array( i_array ).av_model, azel_mat, p );
	
      end
      
    end
    
    % Sanity check

    if ~all( size( speechpix_mat ) == im_size( 1:2 ) )
      error( [ mfilename ': insanity #123.' ] );
    end

    % Write out speechpix_mat in appropriate grey levels

    output_filename = [ p.output_namehead  'speechpix_mat_' name_list{ a } ];
    output_filename( end-2:end ) = 'pgm';

    write_speechpix_mat( speechpix_mat, n_array, output_filename, p );
    
    % ( 2.5 ) Update the improved V background model, using both V measurements and speechpix_mat
    % -> the only modification w.r.t. Stauffer PAMI 2000
    % is that weight update is not done on speech pixels.
    %
    % This way we hope to keep a person longer in the foreground,
    % as long as she speaks.
    
    v_model = update_v_model( v_model, im, p, speechpix_mat, n_array );

    % ( 2.6 ) Take a background/foreground decision based on the improved V background model
    % This function also write it out to a file, and returns the result
    
    output_filename = [ p.output_namehead  name_list{ a } ];

    im_classif2 = decide_bgfg( v_model, im, output_filename, p );
   
    
    if p.verbose
      chrono = chrono_check( chrono, a );
    end

  end  % Of loop through images  

  if p.verbose
    chrono_stop( chrono );
  end

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % ( end ) Write out models parameters, as obtained
  % at the last video frame.
  
  % Save the model parameters
  save( output_model_filename, 'array', 'v_model' );
  if p.verbose
    disp( [ mfilename ' wrote out "' output_model_filename '".' ] );
  end
  


  if p.verbose
    disp( [ mfilename ': done.' ] );
  end

  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function timecode_list = load_timecodes( input_timecodefile, videoframerate )

  if nargin < 2
    error( [ mfilename '.load_timecodes() needs 2 input arguments.' ] );
  end
  
  fid = fopen( input_timecodefile, 'rt' );
  str = fread( fid, +Inf, 'char' );
  fclose( fid );
  str = char( str(:).' );

  % Detect and split the relevant lines

  [ s, f, t_list ] = regexp( str, '->[\s]+([\S]+.ppm)[\s]+(\d+:\d+:\d+\.\d+)' );
  
  n = numel( t_list );

  timecode_list = cell( 2, n );

  for a = 1:n
    
    t = t_list{ a };
    
    % image name
    timecode_list{ 1, a } = str( t( 1, 1 ):t( 1, 2 ) );
    
    % numerical timecode
    timecode_list{ 2, a } = timecode_to_sec( str( t( 2, 1 ):t( 2, 2 ) ), videoframerate );
    
  end
  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function audioloc_mat = load_audio( input_arrayfile, audiostart_videotime )

  if nargin < 2
    error( [ mfilename '.load_audio() needs 2 input arguments.' ] );
  end
  
  load( input_arrayfile, 'result' );
  
  if numel( result.list ) ~= 1
    % Just in case...
    error( [ mfilename ' can only deal with audio results for a single recording.' ] );
  end
  
  % For each location estimate, extract its videotime
  
  audioloc_audiotime = result.list.framelength_sec / 2 + ( result.list.sphloc_mat( 1, : ) - 1 ) * result.list.frameshift_sec;
  
  audioloc_videotime = audiostart_videotime + audioloc_audiotime;
  
  % Store it, along with azimuth and elevation

  audioloc_mat = [ audioloc_videotime(:).'; result.list.sphloc_mat( 3:4, : ) ];
  


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function av_model = update_av_model( av_model, azel_mat, bg, im_size, p )

  if nargin < 5
    error( [ mfilename '.update_av_model()' ] );
  end

  % Also list the foreground pixels, as per "bg"
  
  fg_ind = find( bg > 0 );
  [ fg_row,  fg_col ] = find( bg > 0 );

  fg_ind = fg_ind(:);
  fg_row = fg_row(:);
  fg_col = fg_col(:);

  n_fg = numel( fg_ind );
  
  for b = 1:size( azel_mat, 2 )

    disp( sprintf( '    audio: %d', b ) ); % XXX debug

    % For each audio location estimate we update the ( A | V ) models of the foreground pixels
    % ( foreground as per "bg" )

    if isempty( av_model )

      % Initialization

      clear av_model
      
      av_model.im_size = im_size;

      av_model.mu      = zeros( [ im_size( 1:2 )  2  p.a_K ] );  % "2" for azimuth and elevation

      tmp = p.a_sigma_init( 1:2 );
      av_model.sigma   = repmat( permute( tmp(:), [ 3 2 1 ] ), [ av_model.im_size( 1:2 ) 1 p.a_K ] );

      av_model.w       = repmat( 1 / p.a_K,      [ av_model.im_size( 1:2 ) 1 p.a_K ] );

      % Smarter initialization
      % -> start with zero component,
      % -> give full weight (1.0) to the very first measurement
      % -> any new measurement that does not fit will lead to a new component
      % -> etc. until p.a_K components have been created (i.e. have non-zero weight)
      %
      % All this is implemented by the next line:
      av_model.w(:) = 0;

      % For convenience when plotting
      av_model.mu(:) = NaN;
      av_model.sigma(:) = NaN;
      
      % Also store the parameters
      
      av_model.par = p;
      
    end

    if ~all( im_size == av_model.im_size )
      error( [ mfilename ': the image dimensionality is not supposed to change over time.' ] );
    end
    
    % Pick the location estimate (az and el)

    azel = azel_mat( 1:2, b );

    % Update of the ( A | V ) model, for each foreground pixel
    
    % First, test for a match, for each pixel and each distribution

    tmp_azel = repmat( permute( azel(:), [ 3 2 1 ] ), [ n_fg 1 1 p.a_K ] );

    ind1 = repmat( fg_row(:), [ 1 1 2 p.a_K ] );
    ind2 = repmat( fg_col(:), [ 1 1 2 p.a_K ] );
    ind3 = repmat( permute( ( 1:2 ).', [ 3 2 1 ] ), [ n_fg 1 1 p.a_K ] );
    ind4 = repmat( permute( ( 1:p.a_K ).', [ 4 3 2 1 ] ), [ n_fg 1 2 1 ] ); 
    ind  = sub2ind( size( av_model.mu ), ind1, ind2, ind3, ind4 );
    
    tmp_mu    = av_model.mu( ind );
    tmp_sigma = av_model.sigma( ind );
    
    delta = max( abs( -pi + mod( pi + tmp_mu - tmp_azel , 2 * pi ) ) ./ ( eps + tmp_sigma ), [], 3 );
    
    % List pixels that have respectively zero match and at least one match
    
    [ delta, sort_ind ] = sort( delta, 4 );
    
    M      = sparse( delta( :, :, :, 1 ) < p.a_Z );
    M0_ind = fg_ind( find( ~M ) );
    M1_ind = fg_ind( find(  M ) );
    
    % In case of no match, replace the least probable distribution with a new one,
    % centered around the current pixel value, and with a large variance.
    
    if ~isempty( M0_ind )
      
      [ ind1, ind2 ] = ind2sub( av_model.im_size, M0_ind );
      
      ind1 = repmat( ind1(:).', 2, 1 );  % "2" for ( az, el )
      ind2 = repmat( ind2(:).', 2, 1 );  % "2" for ( az, el )
      
      ind3 = repmat( ( 1:2 ).', 1, size( ind1, 2 ) );
      
      ind4 = repmat( p.a_K, size( ind1 ) ); % p.a_K: "Least probable" background distribution
      
      mu_ind = sub2ind( size( av_model.mu ), ind1, ind2, ind3, ind4 );
      
      % New distribution: centered around the current azel value
      
      av_model.mu( mu_ind ) = repmat( azel(:), 1, size( mu_ind, 2 ) );
      
      % New distribution: large variance
      
      av_model.sigma( mu_ind ) = repmat( p.a_sigma_init(:), 1, size( mu_ind, 2 ) );

      % New distribution: small weight

      w_ind = sub2ind( size( av_model.w ), ind1( 1, : ), ind2( 1, : ), ind3( 1, : ), ind4( 1, : ) );

      av_model.w( w_ind ) = p.a_w_init;
      
    end  % of update of ( A | V ) on foreground pixels that don't have any match
    
    % Update the best-matching distribution with online K-means,
    % and others with decay.
    
    % Part 1/2 of the weight update

    ind1 = repmat( fg_row(:), [ 1 1 1 p.a_K ] );
    ind2 = repmat( fg_col(:), [ 1 1 1 p.a_K ] );
    ind3 = ones( [ n_fg 1 1 p.a_K ] );
    ind4 = repmat( permute( ( 1:p.a_K ).', [ 4 3 2 1 ] ), [ n_fg 1 1 1 ] );
    w_ind  = sub2ind( size( av_model.w ), ind1, ind2, ind3, ind4 );

    av_model.w( w_ind ) = ( 1 - p.a_alpha ) * av_model.w( w_ind );

    % Update the means and variances, when relevant
    
    if ~isempty( M1_ind )
      
      [ ind1, ind2 ] = ind2sub( av_model.im_size, M1_ind );
      
      ind1 = repmat( ind1(:).', 2, 1 );  % "2" for ( az, el )
      ind2 = repmat( ind2(:).', 2, 1 );  % "2" for ( az, el )
      
      ind3 = repmat( ( 1:2 ).', 1, size( ind1, 2 ) );
      
      tmp  = sort_ind( :, 1, 1, 1 );  % Best match
      tmp  = tmp( find( M ) );
      ind4 = repmat( tmp(:).', 2, 1 );
      
      mu_ind = sub2ind( size( av_model.mu ), ind1, ind2, ind3, ind4 );
      
      w_ind = sub2ind( size( av_model.w ), ind1( 1, : ), ind2( 1, : ), ind3( 1, : ), ind4( 1, : ) );

      % Part 2/2 of the weight update
      
      av_model.w( w_ind ) = av_model.w( w_ind ) + p.a_alpha;

      % Mean and variance update
      
      tmp_azel = repmat( azel(:), 1, size( mu_ind, 2 ) );

      if p.a_use_fixed_rho
	
	rho_diag = p.a_rho_factor * p.a_alpha;
	
      else

	% Diagonal 2 x 2 covariance matrices,
	% with the two different variances
	%
	% Evaluate the lkld for the best match Gaussian component.
	% From the lkld, we deduce rho.
	% "diag( sparse(" is just to make later computations more convenient.
	
	d   = 2;

	tmp_delta = abs( -pi + mod( pi + tmp_azel - av_model.mu( mu_ind ), 2 * pi ) );
	tmp = tmp_delta ./ ( eps + av_model.sigma( mu_ind ) );
	tmp = ( ( 2*pi ) .^ ( -d/2 ) ) * ( prod( av_model.sigma( mu_ind ), 1 ) .^ ( -1 ) ) .* exp( -0.5 * sum( tmp .^ 2, 1 ) );
	
	rho_diag  = diag( sparse( p.a_alpha * tmp ) );
	
      end
      
      % Update the means
      
      av_model.mu( mu_ind ) = angle( exp( j * av_model.mu( mu_ind ) ) + exp( j * ( tmp_azel - av_model.mu( mu_ind ) ) ) * rho_diag );
      
      % Update the standard deviations
      
      tmp_delta                = abs( -pi + mod( pi + tmp_azel - av_model.mu( mu_ind ), 2 * pi ) );
      sigma2                   = av_model.sigma( mu_ind ) .^ 2;
      av_model.sigma( mu_ind ) = sqrt( max( p.a_sigma_floor .^ 2, sigma2 + ( tmp_delta .^ 2 - sigma2 ) * rho_diag ) );
      
      % ( leave unchanged the other no-match or worse-match distributions )
      
    end  % Update ( A | V ) model on foreground pixels that have at least one match
    
    % Rescale the weights
    
    ind1 = repmat( fg_row(:), [ 1 1 1 p.a_K ] );
    ind2 = repmat( fg_col(:), [ 1 1 1 p.a_K ] );
    ind3 = ones( [ n_fg 1 1 p.a_K ] );
    ind4 = repmat( permute( ( 1:p.a_K ).', [ 4 3 2 1 ] ), [ n_fg 1 1 1 ] ); 
    w_ind  = sub2ind( size( av_model.w ), ind1, ind2, ind3, ind4 );

    av_model.w( w_ind ) = max( 0, av_model.w( w_ind ) );
    av_model.w( w_ind ) = av_model.w( w_ind ) ./ repmat( sum( av_model.w( w_ind ), 4 ), [ 1 1 1 p.a_K ] );

    % Sort the ( A | V ) models on the foreground pixels
    
    ind1 = repmat( fg_row(:), [ 1 1 2 p.a_K ] );
    ind2 = repmat( fg_col(:), [ 1 1 2 p.a_K ] );
    ind3 = repmat( permute( ( 1:2 ).', [ 3 2 1 ] ), [ n_fg 1 1 p.a_K ] );
    ind4 = repmat( permute( ( 1:p.a_K ).', [ 4 3 2 1 ] ), [ n_fg 1 2 1 ] );

    w_ind  = sub2ind( size( av_model.w ), ind1( :,:,1,: ), ind2( :,:,1,: ), ind3( :,:,1,: ) , ind4( :,:,1,: ) );
    
    sigma_ind  = sub2ind( size( av_model.mu ), ind1, ind2, ind3, ind4 );
    
    [ tmp, sort_ind ] = sort( -av_model.w( w_ind ) ./ ( eps + sqrt( max( 0, prod( av_model.sigma( sigma_ind ), 3 ) ) ) ), 4 );
    
    new_ind4      = repmat( sort_ind, [ 1 1 2 1 ] );

    new_w_ind     = sub2ind( size( av_model.w ), ind1( :,:,1,: ), ind2( :,:,1,: ), ind3( :,:,1,: ) , new_ind4( :,:,1,: ) );

    new_sigma_ind    = sub2ind( size( av_model.mu ), ind1, ind2, ind3, new_ind4 );
    
    av_model.w( w_ind )         = av_model.w( new_w_ind );
    av_model.mu( sigma_ind )    = av_model.mu( new_sigma_ind );
    av_model.sigma( sigma_ind ) = av_model.sigma( new_sigma_ind );
    
  end % of loop through audio location estimates
  
  % Mark down which models correspond to speech, based on the ordering
  % done just above.
  
  tmp = av_model.w ./ ( eps + sqrt( max( 0, prod( av_model.sigma, 3 ) ) ) );

  av_model.is_speech = ( cumsum( cat( 4, zeros( av_model.im_size( 1:2 ) ), tmp( :,:,1,1:end-1 ) ), 4 ) < p.a_T );
  
  tmp = any( av_model.w, 4 );

  av_model.is_speech = av_model.is_speech & repmat( tmp, [ 1 1 1 p.a_K ] );
  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function speechpix_mat = detect_speech_pixels( av_model, azel_mat, p )
  
  if nargin < 3
    error( [ mfilename '.detect_speech_pixels() needs 3 input arguments.' ] );
  end
  
  speechpix_mat = zeros( av_model.im_size( 1:2 ) );

  % For each audio location estimate
  
  for b = 1:size( azel_mat, 2 )

    azel = azel_mat( 1:2, b );

    % Try to match

    tmp_azel = repmat( permute( azel(:), [ 3 2 1 ] ), [ av_model.im_size( 1:2 ) 1 p.a_K ] );

    delta = max( abs( -pi + mod( pi + av_model.mu - tmp_azel , 2 * pi ) ) ./ ( eps + av_model.sigma ), [], 3 );
    
    % Mark any pixel that has at least one speech model matching the azel measurement

    speechpix_mat = speechpix_mat | any( av_model.is_speech & ( delta < p.a_Z ), 4 );
    
  end
  


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function v_model = update_v_model( v_model, im, p, speechpix_mat, n_array )

  if nargin < 5
    error( [ mfilename '.update_v_model() needs 5 input arguments.' ] );
  end
  
  im_size = size( im );
  
  
  if isempty( v_model )
    
    % Initialization
    
    v_model.im_size = im_size;
    
    v_model.mu      = repmat( im,           [ 1 1 1 p.v_K ] );
    v_model.mu( :,:,:,2:end ) = -v_model.mu( :,:,:,2:end );  % Make only one Gaussian very likely, others unlikely
    v_model.sigma   = repmat( p.v_sigma_init, [ v_model.im_size( 1:2 ) 1 p.v_K ] );
    v_model.w       = repmat( 1 / p.v_K,      [ v_model.im_size( 1:2 ) 1 p.v_K ] );

    % Also store the parameters

    v_model.par = p;
    
  end
  
  % Online update

  % Test for a match, for each pixel and each distribution
  
  if ~all( im_size == v_model.im_size )
    error( [ mfilename ': the image dimensionality is not supposed to change over time.' ] );
  end
  
  delta = repmat( NaN, [ v_model.im_size( 1:2 )  1  p.v_K ] );
  
  delta = max( abs( repmat( im, [ 1 1 1 p.v_K ] ) - v_model.mu ), [], 3 ) ./ ( eps + v_model.sigma );
  
  if any( isnan( delta(:) ) )
    error( [ mfilename ' detected an unexpected NaN.' ] );
  end

  % List pixels that have respectively zero match and at least one match

  [ delta, sort_ind ] = sort( delta, 4 );
  
  M      = sparse( delta( :, :, :, 1 ) < p.v_Z );
  M0_ind = find( ~M );
  M1_ind = find(  M );
  
  % In case of no match, replace the least probable distribution with a new one,
  % centered around the current pixel value, and with a large variance.

  if ~isempty( M0_ind )
    
    [ ind1, ind2 ] = ind2sub( size( M ), M0_ind );
    
    ind1 = repmat( ind1(:).', v_model.im_size( 3 ), 1 );
    ind2 = repmat( ind2(:).', v_model.im_size( 3 ), 1 );

    ind3 = repmat( ( 1:v_model.im_size( 3 ) ).', 1, size( ind1, 2 ) );
    
    ind4 = repmat( p.v_K, size( ind1 ) ); % p.v_K: "Least probable" background distribution
    
    mu_ind = sub2ind( size( v_model.mu ), ind1, ind2, ind3, ind4 );
    im_ind = sub2ind( size( im ),       ind1, ind2, ind3 );

    % New distribution: centered around the current pixel value

    v_model.mu( mu_ind ) = im( im_ind );

    % New distribution: large variance
    
    sigma_ind = sub2ind( size( v_model.sigma ), ind1( 1, : ), ind2( 1, : ), ones( 1, size( ind1, 2 ) ), ind4( 1, : ) );
    
    v_model.sigma( sigma_ind ) = p.v_sigma_init;
    
    % New distribution: small weight

    w_ind = sigma_ind;

    v_model.w( w_ind ) = p.v_w_init;
    
  end

  % Update the best-matching distribution with online K-means,
  % and others with decay.

  M4 = repmat( permute( ( 1:p.v_K ).', [ 4 3 2 1 ] ), [ v_model.im_size( 1:2 ) 1 1 ] );
  M4 = ( M4 == repmat( sort_ind( :,:,1,1 ), [ 1 1 1 p.v_K ] ) );
  
  % Update the weights ON SILENT PIXELS ONLY
  
  w_ind = find( repmat( speechpix_mat == 0, [ 1 1 1 p.v_K ] ) );
  
  v_model.w( w_ind ) = full( max( 0, v_model.w( w_ind ) + p.v_alpha * ( M4( w_ind ) - v_model.w( w_ind ) ) ) );
  
  % Update the means and variances, when relevant
  
  if ~isempty( M1_ind )
    
    [ ind1, ind2 ] = ind2sub( size( M ), M1_ind );
    
    ind1 = repmat( ind1(:).', v_model.im_size( 3 ), 1 );
    ind2 = repmat( ind2(:).', v_model.im_size( 3 ), 1 );
    
    ind3 = repmat( ( 1:v_model.im_size( 3 ) ).', 1, size( ind1, 2 ) );
    
    tmp  = sort_ind( :, :, 1, 1 );  % Best match
    tmp  = tmp( M1_ind );
    ind4 = repmat( tmp(:).', v_model.im_size( 3 ), 1 );
    
    mu_ind    = sub2ind( size( v_model.mu ), ind1, ind2, ind3, ind4 );
    im_ind    = sub2ind( size( im ),       ind1, ind2, ind3 );

    sigma_ind = sub2ind( size( v_model.sigma ), ind1( 1, : ), ind2( 1, : ), ones( 1, size( ind1, 2 ) ), ind4( 1, : ) );

    if p.v_use_fixed_rho

      rho_diag = p.v_rho_factor * p.v_alpha;

    else

      % Diagonal covariance matrices,
      % with the same variance for all components
      %
      % Evaluate the lkld for the best match Gaussian component.
      % From the lkld, we deduce rho.
      % "diag( sparse(" is just to make later computations more convenient.
      
      d   = v_model.im_size( 3 );
      tmp = ( ( 2*pi ) .^ ( -d/2 ) ) * ( v_model.sigma( sigma_ind ) .^ ( -d ) ) .* exp( -1 ./ ( 2 * ( v_model.sigma( sigma_ind ) .^ 2 ) ) .* sum( ( im( im_ind ) - v_model.mu( mu_ind ) ) .^ 2, 1 ) );
      
      rho_diag  = diag( sparse( p.v_alpha * tmp ) );
      
    end
    
    % Update the means
    
    v_model.mu( mu_ind ) = v_model.mu( mu_ind ) + ( im( im_ind ) - v_model.mu( mu_ind ) ) * rho_diag;
    
    % Update the standard deviations
    
    sigma2                   = v_model.sigma( sigma_ind ) .^ 2;
    v_model.sigma( sigma_ind ) = sqrt( sigma2 + ( sum( ( im( im_ind ) - v_model.mu( mu_ind ) ) .^ 2, 1 ) - sigma2 ) * rho_diag );
    
    % ( leave unchanged the other no-match or worse-match distributions )
    
  end
  
  % Make sure all weights sum to one
  
  v_model.w = max( 0, v_model.w );

  v_model.w = v_model.w ./ ( eps + repmat( sum( v_model.w, 4 ), [ 1 1 1 p.v_K ] ) );
  
  % ( 2.3 ) Classify each pixel between foreground (1) and background (0)
  
  % ( 2.3.1 ) Order the Gaussians by decreasing ( w_k / sigma_k )
  % According to the PAMI article, a large ( w / sigma ) means background,
  % and a small ( w / sigma ) means foreground.
  
  [ tmp, sort_ind ] = sort( -v_model.w ./ ( eps + v_model.sigma ), 4 );

  % Sort mu
  
  ind1 = repmat( ( 1:v_model.im_size( 1 ) ).', [ 1 v_model.im_size( 2 ) v_model.im_size( 3 ) p.v_K ] );
  ind2 = repmat( ( 1:v_model.im_size( 2 ) ),   [ v_model.im_size( 1 ) 1 v_model.im_size( 3 ) p.v_K ] );
  ind3 = repmat( permute( 1:v_model.im_size( 3 ), [ 1 3 2 4 ] ), [ v_model.im_size( 1:2 ) 1 p.v_K ] );
  ind4 = repmat( sort_ind, [ 1 1 v_model.im_size( 3 ) 1 ] );
  
  v_model.mu = v_model.mu( sub2ind( size( v_model.mu ), ind1, ind2, ind3, ind4 ) );

  % Sort sigma and w

  ind1 = repmat( ( 1:v_model.im_size( 1 ) ).', [ 1 v_model.im_size( 2 ) 1 p.v_K ] );
  ind2 = repmat( ( 1:v_model.im_size( 2 ) ),   [ v_model.im_size( 1 ) 1 1 p.v_K ] );
  ind3 = ones( [ v_model.im_size( 1:2 ) 1 p.v_K ] );
  ind4 = sort_ind;
  
  ind = sub2ind( size( v_model.w ), ind1, ind2, ind3, ind4 );
  
  v_model.sigma = v_model.sigma( ind );
  v_model.w     = v_model.w( ind );
  
  % ( 2.3.2 ) Determine which Gaussians belong to the background
  
  v_model.is_bg = cat( 4, zeros( v_model.im_size( 1:2 ) ), cumsum( v_model.w( :, :, :, 1:end-1 ) , 4 ) ) < p.v_T;
  



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function im_classif2 = decide_bgfg( v_model, im, output_filename, p )

  if nargin < 4
    error( [ mfilename '.decide_bgfg() needs 4 input arguments.' ] );
  end
  
  
  % Evaluate the matches

  delta = repmat( NaN, [ v_model.im_size( 1:2 )  1  p.v_K ] );
  
  delta = max( abs( repmat( im, [ 1 1 1 p.v_K ] ) - v_model.mu ), [], 3 ) ./ ( eps + v_model.sigma );
  
  if any( isnan( delta(:) ) )
    error( [ mfilename ' detected an unexpected NaN.' ] );
  end
  
  % For each pixel, count the number of matched background Gaussians (may be zero)
  
  n_bg_match = sum( ( delta < p.v_Z ) & v_model.is_bg, 4 );

  % Classify each pixel:
  %   0 iff background.
  %   255 iff foreground.
  
  im_classif2 = 255 * ( n_bg_match < 1 );
  
  % ( 2.4 ) Write out the result of the classification


  if exist( output_filename, 'file' )

    if p.verbose
      disp( [ mfilename '.decide_bgfg(): output file already exists. Skipped.' ] );
      disp( [ '"' output_filename '"' ] );
    end

  else
    
    % Make sure that the output directory exists
    output_path = fileparts( output_filename );
    if ~exist( output_path, 'dir' )
      system( [ 'mkdir -p ' output_path ] );
      if p.verbose
	disp( ' ' );
	disp( [ mfilename '.decide_bgfg(): created output directory "' output_path '".' ] );
      end
    end

    % Save the binary pattern
    SauvePgm( output_filename, im_classif2 );
    if p.verbose
      disp( [ mfilename '.decide_bgfg(): wrote out "' output_filename '".' ] );
    end

  end
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function write_speechpix_mat( speechpix_mat, n_array, output_filename, p )

  if nargin < 4
    error( [ mfilename '.write_speechpix_mat() needs 4 input arguments.' ] );
  end

  % Transform to gray levels

  g_lev = max( 0, min( 255, round( speechpix_mat / ( ( 2 .^ n_array ) - 1 ) * 255 ) ) );

  % Write out
  
  if exist( output_filename, 'file' )
    
    if p.verbose
      disp( [ mfilename 'write_speechpix_mat(): output file already exists. Skipped.' ] );
      disp( [ '"' output_filename '"' ] );
    end
    
  else
    
    % Make sure that the output directory exists
    output_path = fileparts( output_filename );
    
    if ~exist( output_path, 'dir' )
    
      system( [ 'mkdir -p ' output_path ] );
      if p.verbose
	disp( ' ' );
	disp( [ mfilename 'write_speechpix_mat(): created output directory "' output_path '".' ] );
      end

    end

    % Save the binary pattern
    SauvePgm( output_filename, g_lev );
    if p.verbose
      disp( [ mfilename 'write_speechpix_mat(): wrote out "' output_filename '".' ] );
    end

  end
